@@ -3,7 +3,7 @@
 # Copyright (c) 1999 by Megginson Technologies.
 # No warranty.  Commercial and non-commercial use freely permitted.
 #
-# $Id: Writer.pm,v 0.3 1999/04/25 13:44:50 david Exp david $
+# $Id: Writer.pm,v 1.23 2004/03/06 22:41:03 josephw Exp $
 ########################################################################
 
 package XML::Writer;
@@ -13,9 +13,8 @@ require 5.004;
 use strict;
 use vars qw($VERSION);
 use Carp;
-use IO;
-
-$VERSION = "0.3";
+use IO::Handle;
+$VERSION = "0.500";
 
 
 
@@ -34,34 +33,42 @@ $VERSION = "0.3";
 sub new {
   my ($class, %params) = (@_);
 
-				# If the user wants namespaces,
-				# intercept the request here; it will
-				# come back to this constructor
-				# from within XML::Writer::Namespaces::new()
+                                # If the user wants namespaces,
+                                # intercept the request here; it will
+                                # come back to this constructor
+                                # from within XML::Writer::Namespaces::new()
   if ($params{NAMESPACES}) {
     delete $params{NAMESPACES};
     return new XML::Writer::Namespaces(%params);
   }
 
-				# Set up $self and basic parameters
+                                # Set up $self and basic parameters
   my $self;
   my $output;
   my $unsafe = $params{UNSAFE};
   my $newlines = $params{NEWLINES};
+  my $dataMode = $params{DATA_MODE};
+  my $dataIndent = $params{DATA_INDENT} || 0;
 
-				# If the NEWLINES parameter is specified,
-				# set the $nl variable appropriately
+                                # If the NEWLINES parameter is specified,
+                                # set the $nl variable appropriately
   my $nl = '';
   if ($newlines) {
     $nl = "\n";
   }
 
 
-				# Parse variables
+                                # Parse variables
   my @elementStack = ();
   my $elementLevel = 0;
   my %seen = ();
 
+  my $hasData = 0;
+  my @hasDataStack = ();
+  my $hasElement = 0;
+  my @hasElementStack = ();
+  my $hasHeading = 0; # Does this document have anything before the first element?
+
   #
   # Private method to show attributes.
   #
@@ -75,9 +82,9 @@ sub new {
     }
   };
 
-				# Method implementations: the SAFE_
-				# versions perform error checking
-				# and then call the regular ones.
+                                # Method implementations: the SAFE_
+                                # versions perform error checking
+                                # and then call the regular ones.
   my $end = sub {
     $output->print("\n");
   };
@@ -97,6 +104,13 @@ sub new {
 
   my $xmlDecl = sub {
     my ($encoding, $standalone) = (@_);
+    if ($standalone && $standalone ne 'no') {
+      $standalone = 'yes';
+    }
+
+    # This line is questionable, but changing current behaviour
+    # may be a bad idea. There seems to be a mismatch with the
+    # documentation, though.
     $encoding = "UTF-8" unless $encoding;
     $output->print("<?xml version=\"1.0\"");
     if ($encoding) {
@@ -106,6 +120,7 @@ sub new {
       $output->print(" standalone=\"$standalone\"");
     }
     $output->print("?>\n");
+    $hasHeading = 1;
   };
 
   my $SAFE_xmlDecl = sub {
@@ -127,18 +142,21 @@ sub new {
     }
     if ($elementLevel == 0) {
       $output->print("\n");
+      $hasHeading = 1;
     }
   };
 
   my $SAFE_pi = sub {
     my ($name, $data) = (@_);
     $seen{ANYTHING} = 1;
-    if ($name =~ /xml/i) {
+    if (($name =~ /^xml/i) && ($name !~ /^xml-stylesheet$/i)) {
       carp("Processing instruction target begins with 'xml'");
     } 
 
-    if ($name =~ /\?\>/ || $data =~ /\?\>/) {
+    if ($name =~ /\?\>/ || (defined($data) && $data =~ /\?\>/)) {
       croak("Processing instruction may not contain '?>'");
+    } elsif ($name =~ /\s/) {
+      croak("Processing instruction name may not contain whitespace");
     } else {
       &{$pi};
     }
@@ -149,6 +167,7 @@ sub new {
     $output->print("<!-- $data -->");
     if ($elementLevel == 0) {
       $output->print("\n");
+      $hasHeading = 1;
     }
   };
 
@@ -170,11 +189,15 @@ sub new {
     my ($name, $publicId, $systemId) = (@_);
     $output->print("<!DOCTYPE $name");
     if ($publicId) {
+      unless ($systemId) {
+        croak("A DOCTYPE declaration with a public ID must also have a system ID");
+      }
       $output->print(" PUBLIC \"$publicId\" \"$systemId\"");
     } elsif ($systemId) {
       $output->print(" SYSTEM \"$systemId\"");
     }
     $output->print(">\n");
+    $hasHeading = 1;
   };
 
   my $SAFE_doctype = sub {
@@ -192,11 +215,22 @@ sub new {
 
   my $startTag = sub {
     my $name = $_[0];
+    if ($dataMode && ($hasHeading || $elementLevel)) {
+      $output->print("\n");
+      $output->print(" " x ($elementLevel * $dataIndent));
+    }
     $elementLevel++;
     push @elementStack, $name;
     $output->print("<$name");
     &{$showAttributes}(\@_);
     $output->print("$nl>");
+    if ($dataMode) {
+      $hasElement = 1;
+      push @hasDataStack, $hasData;
+      $hasData = 0;
+      push @hasElementStack, $hasElement;
+      $hasElement = 0;
+    }
   };
 
   my $SAFE_startTag = sub {
@@ -208,8 +242,10 @@ sub new {
       croak("Attempt to insert start tag after close of document element");
     } elsif ($elementLevel == 0 && $seen{DOCTYPE} && $name ne $seen{DOCTYPE}) {
       croak("Document element is \"$name\", but DOCTYPE is \""
-	    . $seen{DOCTYPE}
-	    . "\"");
+            . $seen{DOCTYPE}
+            . "\"");
+    } elsif ($dataMode && $hasData) {
+      croak("Mixed content not allowed in data mode: element $name");
     } else {
       $seen{ANYTHING} = 1;
       $seen{ELEMENT} = 1;
@@ -219,9 +255,16 @@ sub new {
 
   my $emptyTag = sub {
     my $name = $_[0];
+    if ($dataMode && ($hasHeading || $elementLevel)) {
+      $output->print("\n");
+      $output->print(" " x ($elementLevel * $dataIndent));
+    }
     $output->print("<$name");
     &{$showAttributes}(\@_);
     $output->print("$nl />");
+    if ($dataMode) {
+      $hasElement = 1;
+    }
   };
 
   my $SAFE_emptyTag = sub {
@@ -233,8 +276,10 @@ sub new {
       croak("Attempt to insert empty tag after close of document element");
     } elsif ($elementLevel == 0 && $seen{DOCTYPE} && $name ne $seen{DOCTYPE}) {
       croak("Document element is \"$name\", but DOCTYPE is \""
-	    . $seen{DOCTYPE}
-	    . "\"");
+            . $seen{DOCTYPE}
+            . "\"");
+    } elsif ($dataMode && $hasData) {
+      croak("Mixed content not allowed in data mode: element $name");
     } else {
       $seen{ANYTHING} = 1;
       $seen{ELEMENT} = 1;
@@ -247,7 +292,15 @@ sub new {
     my $currentName = pop @elementStack;
     $name = $currentName unless $name;
     $elementLevel--;
+    if ($dataMode && $hasElement) {
+      $output->print("\n");
+      $output->print(" " x ($elementLevel * $dataIndent));
+    }
     $output->print("</$name$nl>");
+    if ($dataMode) {
+      $hasData = pop @hasDataStack;
+      $hasElement = pop @hasElementStack;
+    }
   };
 
   my $SAFE_endTag = sub {
@@ -270,48 +323,84 @@ sub new {
       $data =~ s/\>/\&gt\;/g;
     }
     $output->print($data);
+    $hasData = 1;
   };
 
   my $SAFE_characters = sub {
     if ($elementLevel < 1) {
       croak("Attempt to insert characters outside of document element");
+    } elsif ($dataMode && $hasElement) {
+      croak("Mixed content not allowed in data mode: characters");
     } else {
       &{$characters};
     }
   };
 
-  
-				# Assign the correct closures based on
-				# the UNSAFE parameter
+  my $raw = sub {
+    $output->print($_[0]);
+    # Don't set $hasData or any other information: we know nothing
+    # about what was just written.
+    #
+  };
+
+  my $SAFE_raw = sub {
+  	croak('raw() is only available when UNSAFE is set');
+  };
+
+  my $cdata = sub {
+      my $data = $_[0];
+      $data    =~ s/\]\]>/\]\]\]\]><!\[CDATA\[>/g;
+      $output->print("<![CDATA[$data]]>");
+      $hasData = 1;
+  };
+
+  my $SAFE_cdata = sub {
+    if ($elementLevel < 1) {
+      croak("Attempt to insert characters outside of document element");
+    } elsif ($dataMode && $hasElement) {
+      croak("Mixed content not allowed in data mode: characters");
+    } else {
+      &{$cdata};
+    }
+  };
+
+                                # Assign the correct closures based on
+                                # the UNSAFE parameter
   if ($unsafe) {
-    $self = {END => $end,
-	     XMLDECL => $xmlDecl,
-	     PI => $pi,
-	     COMMENT => $comment,
-	     DOCTYPE => $doctype,
-	     STARTTAG => $startTag,
-	     EMPTYTAG => $emptyTag,
-	     ENDTAG => $endTag,
-	     CHARACTERS => $characters};
+    $self = {'END' => $end,
+             'XMLDECL' => $xmlDecl,
+             'PI' => $pi,
+             'COMMENT' => $comment,
+             'DOCTYPE' => $doctype,
+             'STARTTAG' => $startTag,
+             'EMPTYTAG' => $emptyTag,
+             'ENDTAG' => $endTag,
+             'CHARACTERS' => $characters,
+             'RAW' => $raw,
+             'CDATA' => $cdata
+            };
   } else {
-    $self = {END => $SAFE_end,
-	     XMLDECL => $SAFE_xmlDecl,
-	     PI => $SAFE_pi,
-	     COMMENT => $SAFE_comment,
-	     DOCTYPE => $SAFE_doctype,
-	     STARTTAG => $SAFE_startTag,
-	     EMPTYTAG => $SAFE_emptyTag,
-	     ENDTAG => $SAFE_endTag,
-	     CHARACTERS => $SAFE_characters};
+    $self = {'END' => $SAFE_end,
+             'XMLDECL' => $SAFE_xmlDecl,
+             'PI' => $SAFE_pi,
+             'COMMENT' => $SAFE_comment,
+             'DOCTYPE' => $SAFE_doctype,
+             'STARTTAG' => $SAFE_startTag,
+             'EMPTYTAG' => $SAFE_emptyTag,
+             'ENDTAG' => $SAFE_endTag,
+             'CHARACTERS' => $SAFE_characters,
+             'RAW' => $SAFE_raw,               # This will intentionally fail
+             'CDATA' => $SAFE_cdata
+            };
   }
 
-				# Query methods
-  $self->{IN_ELEMENT} = sub {
+                                # Query methods
+  $self->{'IN_ELEMENT'} = sub {
     my ($ancestor) = (@_);
     return $elementStack[$#elementStack] eq $ancestor;
   };
 
-  $self->{WITHIN_ELEMENT} = sub {
+  $self->{'WITHIN_ELEMENT'} = sub {
     my ($ancestor) = (@_);
     my $el;
     foreach $el (@elementStack) {
@@ -320,36 +409,52 @@ sub new {
     return 0;
   };
 
-  $self->{CURRENT_ELEMENT} = sub {
+  $self->{'CURRENT_ELEMENT'} = sub {
     return $elementStack[$#elementStack];
   };
 
-  $self->{ANCESTOR} = sub {
+  $self->{'ANCESTOR'} = sub {
     my ($n) = (@_);
     return $elementStack[$#elementStack-$n];
   };
 
-				# Set and get the output destination.
-  $self->{GETOUTPUT} = sub {
+                                # Set and get the output destination.
+  $self->{'GETOUTPUT'} = sub {
     return $output;
   };
 
-  $self->{SETOUTPUT} = sub {
+  $self->{'SETOUTPUT'} = sub {
     my $newOutput = $_[0];
-				# If there is no OUTPUT parameter,
-				# use standard output
+                                # If there is no OUTPUT parameter,
+                                # use standard output
     unless ($newOutput) {
       $newOutput = new IO::Handle();
       $newOutput->fdopen(fileno(STDOUT), "w") ||
-	croak("Cannot write to standard output");
+        croak("Cannot write to standard output: $!");
     }
     $output = $newOutput;
   };
 
-				# Set the output.
-  &{$self->{SETOUTPUT}}($params{OUTPUT});
+  $self->{'SETDATAMODE'} = sub {
+    $dataMode = $_[0];
+  };
 
-				# Return the blessed object.
+  $self->{'GETDATAMODE'} = sub {
+    return $dataMode;
+  };
+
+  $self->{'SETDATAINDENT'} = sub {
+    $dataIndent = $_[0];
+  };
+
+  $self->{'GETDATAINDENT'} = sub {
+    return $dataIndent;
+  };
+
+                                # Set the output.
+  &{$self->{'SETOUTPUT'}}($params{'OUTPUT'});
+
+                                # Return the blessed object.
   return bless $self, $class;
 }
 
@@ -423,6 +528,26 @@ sub endTag {
   &{$self->{ENDTAG}};
 }
 
+#
+# Write a simple data element.
+#
+sub dataElement {
+  my ($self, $name, $data, %atts) = (@_);
+  $self->startTag($name, %atts);
+  $self->characters($data);
+  $self->endTag($name);
+}
+
+#
+# Write a simple CDATA element.
+#
+sub cdataElement {
+    my ($self, $name, $data, %atts) = (@_);
+    $self->startTag($name, %atts);
+    $self->cdata($data);
+    $self->endTag($name);
+}
+
 #
 # Write character data.
 #
@@ -431,6 +556,22 @@ sub characters {
   &{$self->{CHARACTERS}};
 }
 
+#
+# Write raw, unquoted, completely unchecked character data.
+#
+sub raw {
+  my $self = shift;
+  &{$self->{RAW}};
+}
+
+#
+# Write CDATA.
+#
+sub cdata {
+    my $self = shift;
+    &{$self->{CDATA}};
+}
+
 #
 # Query the current element.
 #
@@ -480,6 +621,41 @@ sub setOutput {
   return &{$self->{SETOUTPUT}};
 }
 
+#
+# Set the current data mode (true or false).
+#
+sub setDataMode {
+  my $self = shift;
+  return &{$self->{SETDATAMODE}};
+}
+
+
+#
+# Get the current data mode (true or false).
+#
+sub getDataMode {
+  my $self = shift;
+  return &{$self->{GETDATAMODE}};
+}
+
+
+#
+# Set the current data indent step.
+#
+sub setDataIndent {
+  my $self = shift;
+  return &{$self->{SETDATAINDENT}};
+}
+
+
+#
+# Get the current data indent step.
+#
+sub getDataIndent {
+  my $self = shift;
+  return &{$self->{GETDATAINDENT}};
+}
+
 
 #
 # Empty stub.
@@ -508,8 +684,8 @@ sub removePrefix {
 sub _checkAttributes {
   my %anames;
   my $i = 1;
-  while ($_[$i]) {
-    my $name = $_[$i];
+  while ($_[0]->[$i]) {
+    my $name = $_[0]->[$i];
     $i += 2;
     if ($anames{$name}) {
       croak("Two attributes named \"$name\"");
@@ -554,40 +730,48 @@ sub new {
 
   my $unsafe = $params{UNSAFE};
 
-				# Snarf the prefix map, if any, and
-				# note the default prefix.
+                                # Snarf the prefix map, if any, and
+                                # note the default prefix.
   my %prefixMap = ();
   if ($params{PREFIX_MAP}) {
     %prefixMap = (%{$params{PREFIX_MAP}});
     delete $params{PREFIX_MAP};
   }
-  my $defaultPrefix = $prefixMap{''};
-  delete $prefixMap{''};
+  $prefixMap{'http://www.w3.org/XML/1998/namespace'} = 'xml';
 
-				# Generate the reverse map for URIs
+                                # Generate the reverse map for URIs
   my %uriMap = ();
   my $key;
   foreach $key (keys(%prefixMap)) {
     $uriMap{$prefixMap{$key}} = $key;
   }
 
-				# Create an instance of the parent.
+  my $defaultPrefix = $uriMap{''};
+  delete $prefixMap{$defaultPrefix} if ($defaultPrefix);
+
+                                # Create an instance of the parent.
   my $self = new XML::Writer(%params);
 
-				# Snarf the parent's methods that we're
-				# going to override.
+                                # Snarf the parent's methods that we're
+                                # going to override.
   my $OLD_startTag = $self->{STARTTAG};
   my $OLD_emptyTag = $self->{EMPTYTAG};
   my $OLD_endTag = $self->{ENDTAG};
 
-				# State variables
+                                # State variables
   my $prefixCounter = 1;
   my @nsDecls = ();
-  my $nsDecls = {};
+  my $nsDecls = {'http://www.w3.org/XML/1998/namespace' => 'xml'};
   my @nsDefaultDecl = ();
   my $nsDefaultDecl = undef;
   my @nsCopyFlag = ();
   my $nsCopyFlag = 0;
+  my @forcedNSDecls = ();
+
+  if ($params{FORCED_NS_DECLS}) {
+    @forcedNSDecls = @{$params{FORCED_NS_DECLS}};
+    delete $params{FORCED_NS_DECLS};
+  }
 
   #
   # Push the current declaration state.
@@ -629,28 +813,28 @@ sub new {
     my ($uri, $local) = @{$$nameref};
     my $prefix = $prefixMap{$uri};
 
-				# Is this an element name that matches
-				# the default NS?
-    if (!$attFlag && ($uri eq $defaultPrefix)) {
+                                # Is this an element name that matches
+                                # the default NS?
+    if (!$attFlag && $defaultPrefix && ($uri eq $defaultPrefix)) {
       unless ($nsDefaultDecl) {
-	push @{$atts}, 'xmlns';
-	push @{$atts}, $uri;
-	$nsDefaultDecl = 1;
+        push @{$atts}, 'xmlns';
+        push @{$atts}, $uri;
+        $nsDefaultDecl = 1;
       }
       $$nameref = $local;
       
-				# Is there a straight-forward prefix?
+                                # Is there a straight-forward prefix?
     } elsif ($prefix) {
       unless ($nsDecls->{$uri}) {
-				# Copy on write (FIXME: duplicated)
-	unless ($nsCopyFlag) {
-	  $nsCopyFlag = 1;
-	  my %decls = (%{$nsDecls});
-	  $nsDecls = \%decls;
-	}
-	$nsDecls->{$uri} = $prefix;
-	push @{$atts}, "xmlns:$prefix";
-	push @{$atts}, $uri;
+                                # Copy on write (FIXME: duplicated)
+        unless ($nsCopyFlag) {
+          $nsCopyFlag = 1;
+          my %decls = (%{$nsDecls});
+          $nsDecls = \%decls;
+        }
+        $nsDecls->{$uri} = $prefix;
+        push @{$atts}, "xmlns:$prefix";
+        push @{$atts}, $uri;
       }
       $$nameref = "$prefix:$local";
 
@@ -659,9 +843,9 @@ sub new {
       $prefixMap{$uri} = $prefix;
       $uriMap{$prefix} = $uri;
       unless ($nsCopyFlag) {
-	$nsCopyFlag = 1;
-	my %decls = (%{$nsDecls});
-	$nsDecls = \%decls;
+        $nsCopyFlag = 1;
+        my %decls = (%{$nsDecls});
+        $nsDecls = \%decls;
       }
       $nsDecls->{$uri} = $prefix;
       push @{$atts}, "xmlns:$prefix";
@@ -681,10 +865,20 @@ sub new {
     my $i = 1;
     while ($_[0]->[$i]) {
       if (ref($_[0]->[$i]) eq 'ARRAY') {
-	&{$processName}(\$_[0]->[$i], $_[0], 1);
+        &{$processName}(\$_[0]->[$i], $_[0], 1);
       }
       $i += 2;
     }
+
+    # We only do this for the outermost element
+    if (@forcedNSDecls) {
+      foreach (@forcedNSDecls) {
+        my @dummy = ($_, 'dummy');
+        my $d2 = \@dummy;
+        &{$processName}(\$d2, $_[0], 1);
+      }
+      @forcedNSDecls = ();
+    }
   };
 
   #
@@ -733,8 +927,8 @@ sub new {
     my $OLD_pi = $self->{PI};
     $self->{PI} = sub {
       my $target = $_[0];
-      if ($target =~ /:/) {
-	croak "PI target '$target' contains a colon.";
+      if (index($target, ':') >= 0) {
+        croak "PI target '$target' contains a colon.";
       }
       &{$OLD_pi};
     }
@@ -760,7 +954,7 @@ sub new {
   #
   $self->{REMOVEPREFIX} = sub {
     my ($uri) = (@_);
-    if ($defaultPrefix eq $uri) {
+    if ($defaultPrefix && ($defaultPrefix eq $uri)) {
       $defaultPrefix = undef;
     }
     delete $prefixMap{$uri};
@@ -800,32 +994,30 @@ sub _checkNSNames {
   my $i = 1;
   my $name = $names->[0];
 
-				# Check the element name.
+                                # Check the element name.
   if (ref($name) eq 'ARRAY') {
-    if ($name->[1] =~ /:/) {
+    if (index($name->[1], ':') >= 0) {
       croak("Local part of element name '" .
-	    $name->[1] .
-	    "' contains a colon.");
+            $name->[1] .
+            "' contains a colon.");
     }
-  } elsif ($name =~ /:/) {
+  } elsif (index($name, ':') >= 0) {
     croak("Element name '$name' contains a colon.");
   }
 
-				# Check the attribute names.
+                                # Check the attribute names.
   while ($names->[$i]) {
     my $name = $names->[$i];
     if (ref($name) eq 'ARRAY') {
       my $local = $name->[1];
-      if ($local =~ /:/) {
-	croak "Local part of attribute name '$local' contains a colon.";
+      if (index($local, ':') >= 0) {
+        croak "Local part of attribute name '$local' contains a colon.";
       }
     } else {
-      if ($name =~ /^(xmlns|.*:)/) {
-	if ($name =~ /^xmlns/) {
-	  croak "Attribute name '$name' begins with 'xmlns'";
-	} elsif ($name =~ /:/) {
-	  croak "Attribute name '$name' contains ':'";
-	}
+      if ($name =~ /^xmlns/) {
+        croak "Attribute name '$name' begins with 'xmlns'";
+      } elsif (index($name, ':') >= 0) {
+        croak "Attribute name '$name' contains ':'";
       }
     }
     $i += 2;
@@ -847,7 +1039,7 @@ XML::Writer - Perl extension for writing XML documents.
 =head1 SYNOPSIS
 
   use XML::Writer;
-  use IO;
+  use IO::File;
 
   my $output = new IO::File(">output.xml");
 
@@ -877,6 +1069,10 @@ processing mode.  In Namespace mode, the module will generate
 Namespace Declarations itself, and will perform additional checks on
 the output.
 
+Additional support is available for a simplified data mode with no
+mixed content: newlines are automatically inserted around elements and
+elements can optionally be indented based as their nesting level.
+
 
 =head1 METHODS
 
@@ -935,6 +1131,14 @@ generate prefixes of the form "__NS1", "__NS2", etc.
 
 To set the default namespace, use '' for the prefix.
 
+=item FORCED_NS_DECLS
+
+An array reference; if this parameter is present, the document element
+will contain declarations for all the given namespace URIs.
+Declaring namespaces in advance is particularly useful when a large
+number of elements from a namespace are siblings, but don't share a direct
+ancestor from the same namespace.
+
 =item NEWLINES
 
 A true or false value; if this parameter is present and its value is
@@ -951,6 +1155,20 @@ If the parameter is not present, the module will perform the
 well-formedness error checking by default.  Turn off error checking at
 your own risk!
 
+=item DATA_MODE
+
+A true or false value; if this parameter is present and its value is
+true, then the module will enter a special data mode, inserting
+newlines automatically around elements and (unless UNSAFE is also
+specified) reporting an error if any element has both characters and
+elements as content.
+
+=item DATA_INDENT
+
+A numeric value; if this parameter is present, it represents the
+indent step for elements in data mode (it will be ignored when not in
+data mode).
+
 =back
 
 =item end()
@@ -965,7 +1183,9 @@ closed:
 
 Add an XML declaration to the beginning of an XML document.  The
 version will always be "1.0".  If you provide a non-null encoding or
-standalone argument, its value will appear in the declaration.
+standalone argument, its value will appear in the declaration (and
+non-null value for standalone except 'no' will automatically be
+converted to 'yes').
 
   $writer->xmlDecl("UTF-8");
 
@@ -974,7 +1194,7 @@ standalone argument, its value will appear in the declaration.
 Add a DOCTYPE declaration to an XML document.  The declaration must
 appear before the beginning of the root element.  If you provide a
 publicId, you must provide a systemId as well, but you may provide
-just a system ID.
+just a system ID by passing 'undef' for the publicId.
 
   $writer->doctype("html");
 
@@ -1037,6 +1257,15 @@ supply the name of the currently open element:
   $writer->startTag('p');
   $writer->endTag();
 
+=item dataElement($name, $data [, $aname1 => $value1, ...])
+
+Print an entire element containing only character data.  This is
+equivalent to
+
+  $writer->startTag($name [, $aname1 => $value1, ...]);
+  $writer->characters($data);
+  $writer->endTag($name);
+
 =item characters($data)
 
 Add character data to an XML document.  All '<', '>', and '&'
@@ -1049,6 +1278,63 @@ the predefined XML entities:
 You may invoke this method only within the document element
 (i.e. after the first start tag and before the last end tag).
 
+In data mode, you must not use this method to add whitespace between
+elements.
+
+=item raw($data)
+
+Print data completely unquoted and unchecked to the XML document.  For
+example C<raw('<')> will print a literal < character.  This
+necessarily bypasses all well-formedness checking, and is therefore
+only available in unsafe mode.
+
+This can sometimes be useful for printing entities which are defined
+for your XML format but the module doesn't know about, for example
+&nbsp; for XHTML.
+
+=item cdata($data)
+
+As C<characters()> but writes the data quoted in a CDATA section, that
+is, between <![CDATA[ and ]]>.  If the data to be written itself
+contains ]]>, it will be written as several consecutive CDATA
+sections.
+
+=item cdataElement($name, $data [, $aname1 => $value1, ...])
+
+As C<dataElement()> but the element content is written as one or more
+CDATA sections (see C<cdata()>).
+
+=item setOutput($output)
+
+Set the current output destination, as in the OUTPUT parameter for the
+constructor.
+
+=item getOutput()
+
+Return the current output destination, as in the OUTPUT parameter for
+the constructor.
+
+=item setDataMode($mode)
+
+Enable or disable data mode, as in the DATA_MODE parameter for the
+constructor.
+
+=item getDataMode()
+
+Return the current data mode, as in the DATA_MODE parameter for the
+constructor.
+
+=item setDataIndent($step)
+
+Set the indent step for data mode, as in the DATA_INDENT parameter for
+the constructor.
+
+=item getDataIndent()
+
+Return the indent step for data mode, as in the DATA_INDENT parameter
+for the constructor.
+
+
 =back
 
 =head2 Querying XML
